/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <openssl/opensslv.h>
#include <openssl/engine.h>
#include <openssl/evp.h>
#include <openssl/crypto.h>

#include <doca_log.h>

#include "doca_sha_offload_lib.h"

DOCA_LOG_REGISTER(DOCA_SHA_OFFLOAD_ENGINE);

static struct doca_sha_engine_state *engine_state;

static EVP_MD *doca_sha_digest_sha1;
static EVP_MD *doca_sha_digest_sha256;
static EVP_MD *doca_sha_digest_sha512;

#ifdef PRINT_DEBUG_INFO
static uint64_t nb_call_of_init;
static uint64_t nb_call_of_update;
static uint64_t nb_call_of_final;
static uint64_t nb_call_of_copy;
#endif /* PRINT_DEBUG_INFO */

static int debug_level;

static const char * const default_engine_pci_addr = "03:00.0";
static const char *cmdline_engine_pci_addr;

#define DOCA_SHA_ENGINE_CMD_BASE (ENGINE_CMD_BASE)

enum doca_sha_engine_cmd {
#ifdef PRINT_DEBUG_INFO
	DOCA_SHA_ENGINE_DEBUG_LEVEL = DOCA_SHA_ENGINE_CMD_BASE,
	DOCA_SHA_ENGINE_SET_PCI_ADDR,
#else /* PRINT_DEBUG_INFO */
	DOCA_SHA_ENGINE_SET_PCI_ADDR = DOCA_SHA_ENGINE_CMD_BASE,
#endif /* PRINT_DEBUG_INFO */
};

static ENGINE_CMD_DEFN engine_cmd_defns[] = {
#ifdef PRINT_DEBUG_INFO
	{
		.cmd_num = DOCA_SHA_ENGINE_DEBUG_LEVEL,
		.cmd_name = "debug_level",
		.cmd_desc = "enable doca_sha_engine debug print (0, disable; 1, basic; 2, details)",
		.cmd_flags = ENGINE_CMD_FLAG_NUMERIC
	},
#endif /* PRINT_DEBUG_INFO */
	{
		.cmd_num = DOCA_SHA_ENGINE_SET_PCI_ADDR,
		.cmd_name = "set_pci_addr",
		.cmd_desc = "set the pci address of the doca_sha_engine, for example, 03:00.0",
		.cmd_flags = ENGINE_CMD_FLAG_STRING
	},
	{
		0, NULL, NULL, 0
	},
};

static int
engine_ctrl(ENGINE *e, int cmd, long i, void *p, void (*f)(void))
{
	(void)e;
	(void)f;

	switch (cmd) {
#ifdef PRINT_DEBUG_INFO
	case DOCA_SHA_ENGINE_DEBUG_LEVEL:
		debug_level = i;
		if (debug_level > 0)
			DOCA_LOG_DBG("debug level is: %d", debug_level);
		break;
#endif /* PRINT_DEBUG_INFO */
	case DOCA_SHA_ENGINE_SET_PCI_ADDR:
#ifdef PRINT_DEBUG_INFO
		if (debug_level > 0)
			DOCA_LOG_DBG("The default pci_addr = [%s]", default_engine_pci_addr);
#endif /* PRINT_DEBUG_INFO */
		cmdline_engine_pci_addr = (const char *)p;
#ifdef PRINT_DEBUG_INFO
		if (debug_level > 0)
			DOCA_LOG_DBG("The new pci_addr = [%s]", cmdline_engine_pci_addr);
#endif /* PRINT_DEBUG_INFO */
		break;
	default:
		DOCA_LOG_ERR("unsupported ctrl cmd [%d] found", cmd);
		return 0;
	}
	return 1;
}

static int
doca_sha_init(EVP_MD_CTX *ctx)
{
	int ret = 0;
	int sha_type = 0;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0) {
		DOCA_LOG_DBG("enter [%s], NID_sha1 = %d, NID_sha256 = %d, NID_sha512 = %d, recv EVP_MD_CTX_type = %d",
			__func__, NID_sha1, NID_sha256, NID_sha512, EVP_MD_CTX_type(ctx));
	}
#endif /* PRINT_DEBUG_INFO */

	if (ctx == NULL)
		return 0;

	switch (EVP_MD_CTX_type(ctx)) {
	case NID_sha1:
		sha_type = DOCA_SHA_FUNC_TYPE_SHA1;
		break;
	case NID_sha256:
		sha_type = DOCA_SHA_FUNC_TYPE_SHA256;
		break;
	case NID_sha512:
		sha_type = DOCA_SHA_FUNC_TYPE_SHA512;
		break;
	default:
		DOCA_LOG_ERR("unsupported NID [%d]", EVP_MD_CTX_type(ctx));
		return 0;
	}
	ret = doca_sha_init_v1((void *)EVP_MD_CTX_md_data(ctx), sha_type, engine_state, debug_level);

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0)
		nb_call_of_init++;
#endif /* PRINT_DEBUG_INFO */

	return ret;
}

static int
doca_sha_update(EVP_MD_CTX *ctx, const void *data, size_t count)
{
	int ret = 0;

	ret = doca_sha_update_v1((void *)EVP_MD_CTX_md_data(ctx), (unsigned char *)data, count, debug_level);
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0)
		nb_call_of_update++;
#endif /* PRINT_DEBUG_INFO */
	return ret;
}

static int
doca_sha_final(EVP_MD_CTX *ctx, unsigned char *md)
{
	int ret = 0;

	ret = doca_sha_final_v1(md, (void *)EVP_MD_CTX_md_data(ctx), debug_level);
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0)
		nb_call_of_final++;
#endif /* PRINT_DEBUG_INFO */
	return ret;
}

static int
doca_sha_copy(EVP_MD_CTX *to, const EVP_MD_CTX *from)
{
	int ret = 0;

	ret = doca_sha_copy_v1((void *)EVP_MD_CTX_md_data(to),
			    (const void *)EVP_MD_CTX_md_data(from), debug_level);
#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0)
		nb_call_of_copy++;
#endif /* PRINT_DEBUG_INFO */
	return ret;
}

static int
get_digest_info(ENGINE *e, const EVP_MD **digest, const int **nids, int nid)
{
	static const int supported_nids[] = {NID_sha1, NID_sha256, NID_sha512};
	EVP_MD *md = NULL;
	EVP_MD *sha_digest;
	int sha_nid;
	int digest_len;
	int block_len;

#ifdef PRINT_DEBUG_INFO
	if (debug_level) {
		DOCA_LOG_DBG("[%s] enter, NID_sha1 = %d, NID_sha256 = %d, NID_sha512 = %d, the received nid = %d",
			__func__, NID_sha1, NID_sha256, NID_sha512, nid);
	}
#endif /* PRINT_DEBUG_INFO */

	if (digest == NULL) {
		*nids = supported_nids;
		return (sizeof(supported_nids)) / sizeof(supported_nids[0]);
	}

	switch (nid) {
	case NID_sha1:
		sha_digest = doca_sha_digest_sha1;
		sha_nid = NID_sha1;
		digest_len = DOCA_SHA1_DIGEST_LEN;
		block_len = DOCA_SHA1_CBLOCK;
		break;
	case NID_sha256:
		sha_digest = doca_sha_digest_sha256;
		sha_nid = NID_sha256;
		digest_len = DOCA_SHA256_DIGEST_LEN;
		block_len = DOCA_SHA256_CBLOCK;
		break;
	case NID_sha512:
		sha_digest = doca_sha_digest_sha512;
		sha_nid = NID_sha512;
		digest_len = DOCA_SHA512_DIGEST_LEN;
		block_len = DOCA_SHA512_CBLOCK;
		break;
	default:
		return 0;
	}
	if (sha_digest != NULL) {
		*digest = sha_digest;
		return 1;
	}
	md = EVP_MD_meth_new(sha_nid, 0);
	if (md == NULL) {
		DOCA_LOG_ERR("create md failure");
		return 0;
	}
	EVP_MD_meth_set_result_size(md, digest_len);
	EVP_MD_meth_set_input_blocksize(md, block_len);
	EVP_MD_meth_set_app_datasize(md, sizeof(EVP_MD *) + get_size_of_doca_sha_ctx());
	EVP_MD_meth_set_flags(md, EVP_MD_FLAG_ONESHOT);
	EVP_MD_meth_set_init(md, doca_sha_init);
	EVP_MD_meth_set_update(md, doca_sha_update);
	EVP_MD_meth_set_final(md, doca_sha_final);
	EVP_MD_meth_set_copy(md, doca_sha_copy);
	*digest = md;
	switch (nid) {
	case NID_sha1:
		doca_sha_digest_sha1 = md;
		break;
	case NID_sha256:
		doca_sha_digest_sha1 = md;
		break;
	case NID_sha512:
		doca_sha_digest_sha1 = md;
		break;
	}
	return 1;
}

static int
engine_destroy(ENGINE *e)
{
	return doca_sha_engine_destroy(engine_state, debug_level);
}

static int
engine_init(ENGINE *e)
{
	int ret = 0;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0) {
		DOCA_LOG_DBG("default_engine_pci_addr = %s", default_engine_pci_addr);
		DOCA_LOG_DBG("cmdline_engine_pci_addr = %s", cmdline_engine_pci_addr);
	}
	nb_call_of_init = 0;
	nb_call_of_update = 0;
	nb_call_of_final = 0;

	nb_call_of_copy = 0;
#endif /* PRINT_DEBUG_INFO */

	if (doca_sha_engine_create(&engine_state, debug_level) != 0) {
		DOCA_LOG_ERR("doca_sha_engine_create failure");
		return 0;
	}

	if (cmdline_engine_pci_addr != NULL) {
		/*
		 * In the programming mode, user can pass the pci_addr using "set_pci_addr" cmd,
		 * this is the first priority.
		 */
		ret = doca_sha_engine_init(engine_state, cmdline_engine_pci_addr, debug_level);
	} else {
		/*
		 * in the cmdline mode, there is no way to pass the correct pci_addr into the engine.
		 * so user can use makefile to specify the pci_addr
		 */
#ifdef DOCA_ENGINE_PCI_ADDR
#ifdef PRINT_DEBUG_INFO
		if (debug_level > 0)
			DOCA_LOG_DBG("macro DOCA_ENGINE_PCI_ADDR = %s", DOCA_ENGINE_PCI_ADDR);
#endif /* PRINT_DEBUG_INFO */
		if (strlen(DOCA_ENGINE_PCI_ADDR) != 0)
			ret = doca_sha_engine_init(engine_state, DOCA_ENGINE_PCI_ADDR, debug_level);
#endif //DOCA_ENGINE_PCI_ADDR
		if (ret != 1) //the default pci_addr is the lowest priority
			ret = doca_sha_engine_init(engine_state, default_engine_pci_addr, debug_level);
	}
	return ret;
}

static int
engine_finish(ENGINE *e)
{
	int ret = 0;

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0) {
		DOCA_LOG_DBG("nb_call_of_init = %ld, nb_call_of_update = %ld, nb_call_of_final = %ld, nb_call_of_copy = %ld",
			      nb_call_of_init, nb_call_of_update, nb_call_of_final, nb_call_of_copy);
	}
#endif /* PRINT_DEBUG_INFO */

	ret = doca_sha_engine_finish(engine_state, debug_level);

	if (doca_sha_digest_sha1 != NULL) {
		EVP_MD_meth_free(doca_sha_digest_sha1);
		doca_sha_digest_sha1 = NULL;
	}
	if (doca_sha_digest_sha256 != NULL) {
		EVP_MD_meth_free(doca_sha_digest_sha256);
		doca_sha_digest_sha256 = NULL;
	}
	if (doca_sha_digest_sha512 != NULL) {
		EVP_MD_meth_free(doca_sha_digest_sha512);
		doca_sha_digest_sha512 = NULL;
	}
	return ret;
}

static int
bind(ENGINE *e, const char *id)
{
	static const char * const engine_id = "doca_sha_offload_engine";
	static const char * const engine_name = "Openssl SHA offloading engine based on doca_sha";
	int ret = 0;
	static int doca_sha_offload_engine_loaded;

	doca_log_global_level_set(DOCA_LOG_LEVEL_DEBUG);

#ifdef PRINT_DEBUG_INFO
	if (debug_level > 0) {
		DOCA_LOG_DBG("doca_sha engine called with id [%s]", id);
		DOCA_LOG_DBG("doca_sha_offload_engine_loaded = [%d]", doca_sha_offload_engine_loaded);
	}
#endif /* PRINT_DEBUG_INFO */

#ifdef SO_INSTALLED_UNDER_OPENSSEL_DIR
	if (id && strcmp(id, engine_id)) {
		DOCA_LOG_ERR("doca_sha engine called with unexpected id [%s]", id);
		goto end;
	}
#endif
	if (doca_sha_offload_engine_loaded) {
		DOCA_LOG_ERR("doca_sha engine already loaded in openssl");
		goto end;
	}

	doca_sha_offload_engine_loaded = 1;

	if (!ENGINE_set_id(e, engine_id)) {
		DOCA_LOG_ERR("ENGINE_set_id failed");
		goto end;
	}
	if (!ENGINE_set_name(e, engine_name)) {
		DOCA_LOG_ERR("ENGINE_set_name failed");
		goto end;
	}

	if (!ENGINE_set_cmd_defns(e, engine_cmd_defns)) {
		DOCA_LOG_ERR("ENGINE_set_cmd_defns failed");
		goto end;
	}
	if (!ENGINE_set_ctrl_function(e, engine_ctrl)) {
		DOCA_LOG_ERR("ENGINE_set_ctrl_function failed");
		goto end;
	}

	if (!ENGINE_register_digests(e)) {
		DOCA_LOG_ERR("ENGINE_register_digests failed");
		goto end;
	}
	if (!ENGINE_set_digests(e, get_digest_info)) {
		DOCA_LOG_ERR("ENGINE_set_digests failed");
		goto end;
	}

	if (!ENGINE_set_destroy_function(e, engine_destroy)) {
		DOCA_LOG_ERR("ENGINE_set_destroy_function failed");
		goto end;
	}
	if (!ENGINE_set_init_function(e, engine_init)) {
		DOCA_LOG_ERR("ENGINE_set_init_function failed");
		goto end;
	}
	if (!ENGINE_set_finish_function(e, engine_finish)) {
		DOCA_LOG_ERR("ENGINE_set_finish_function failed");
		goto end;
	}

	ret = 1;
end:
	return ret;
}

IMPLEMENT_DYNAMIC_BIND_FN(bind)
IMPLEMENT_DYNAMIC_CHECK_FN()

