/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_log.h>
#include "doca_sha_job_id_generator.h"

DOCA_LOG_REGISTER(DOCA_SHA_OPENSSL_IMPL);

enum id_status {
	ID_STATUS_FREE = 0,
	ID_STATUS_USED,
};

struct doca_sha_id_generator {
	uint8_t *id_array;
	uint32_t max_id;
	uint32_t nb_free_ids;
};

#define NB_ALIGN_BYTE (sizeof(uint64_t))
#define HAS_NOT_FREE_BYTE (0x0101010101010101ul)

struct doca_sha_id_generator *
doca_sha_id_generator_create(uint32_t max_id)
{
	struct doca_sha_id_generator *id_generator;

	if ((max_id & ((NB_ALIGN_BYTE) - 1)) != 0) {
		DOCA_LOG_ERR("max_id for id_generator must be multiple-of-%ld", NB_ALIGN_BYTE);
		return NULL;
	}

	id_generator = calloc(1, sizeof(struct doca_sha_id_generator));
	if (id_generator == NULL) {
		DOCA_LOG_ERR("id allocator calloc failure");
		return NULL;
	}

	id_generator->id_array = calloc(max_id, sizeof(uint8_t));
	if (id_generator->id_array == NULL) {
		DOCA_LOG_ERR("id allocator's id_array calloc failure");
		free(id_generator);
		return NULL;
	}

	memset(id_generator->id_array, ID_STATUS_FREE, sizeof(uint8_t) * (max_id));
	id_generator->max_id = max_id;
	id_generator->nb_free_ids = max_id;
	return id_generator;
}

void
doca_sha_id_generator_destroy(struct doca_sha_id_generator *id_generator)
{
	free(id_generator->id_array);
	free(id_generator);
}

uint32_t
doca_sha_id_generator_get(struct doca_sha_id_generator *id_generator)
{
	uint32_t i = 0;
	uint32_t j = 0;

	if (id_generator->nb_free_ids == 0)
		return DOCA_SHA_ID_GENERATOR_INVALID_ID;

	/* Find the first free 8byte position. */
	for (i = 0; i < (id_generator->max_id / NB_ALIGN_BYTE); i++) {
		if (((uint64_t *)(id_generator->id_array))[i] != HAS_NOT_FREE_BYTE)
			break;
	}
	/* Find the first free byte position. */
	for (j = 0; j < NB_ALIGN_BYTE; j++) {
		if (id_generator->id_array[i * NB_ALIGN_BYTE + j] == ID_STATUS_FREE)
			break;
	}
	id_generator->id_array[i * NB_ALIGN_BYTE + j] = ID_STATUS_USED;
	id_generator->nb_free_ids--;

	return (i * NB_ALIGN_BYTE + j + 1);
}

doca_error_t
doca_sha_id_generator_put(struct doca_sha_id_generator *id_generator, uint32_t id)
{
	if (id < 1) {
		DOCA_LOG_ERR("put id should not < 1");
		return DOCA_ERROR_UNEXPECTED;
	}

	/* Set the position as free. */
	id_generator->id_array[id - 1] = ID_STATUS_FREE;
	id_generator->nb_free_ids++;
	return DOCA_SUCCESS;
}
